try:
    # Django 1.7
    from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation, GenericRel
except ImportError:
    from django.contrib.contenttypes.generic import GenericForeignKey, GenericRelation, GenericRel

__all__ = (
    'GenericForeignKey',
    'GenericRelation',
    'GenericRel',
)



# The timezone support was introduced in Django 1.4, fallback to standard library for 1.3.
try:
    from django.utils.timezone import now, utc
except ImportError:
    # Django < 1.4
    from datetime import datetime
    now = datetime.now
    utc = None   # datetime(..., tzinfo=utc) creates naive datetime this way.

# URLs moved in Django 1.4
try:
    # Django 1.6 requires this
    from django.conf.urls import patterns, url, include
except ImportError:
    # Django 1.3 compatibility, kept in minor release
    from django.conf.urls.defaults import patterns, url, include



import django
from django.conf import settings


class MultiSiteAdminMixin(object):
    """
    Filter the models by ``parent_site``.
    """
    filter_site = True


    def get_queryset(self, request):
        qs = super(MultiSiteAdminMixin, self).get_queryset(request)

        # Admin only shows current site for now,
        # until there is decent filtering for it.
        if self.filter_site:
            qs = qs.parent_site(int(settings.SITE_ID))  # Note: that method can be customized (e.g. SharedContentQuerySet)
        return qs


    # For Django 1.5:
    # Leave for Django 1.6/1.7, so backwards compatibility can be fixed.
    # It will be removed in Django 1.8, so remove it here too to avoid false promises.
    if django.VERSION < (1,8):
        def queryset(self, request):
            qs = super(MultiSiteAdminMixin, self).queryset(request)
            if self.filter_site:
                qs = qs.parent_site(int(settings.SITE_ID))
            return qs



from fluent_utils.django_compat import get_app_label, truncate_name


def get_db_table(module, model_name, meta=None, prefix=None):
    """
    Generate a db_table for a model that's being constructed.
    """
    from django.db import connection

    # This function is used in the metaclasses of django-fluent-pages
    # and django-fluent-contents to generate the db_table with a prefix.
    # This duplicates the logic of Django's Options class
    # to make sure the "db_table" is set in advance while creating the class.
    # Setting the db_table afterwards breaks Django 1.7 migrations,
    # which look at Options.original_attrs['db_table'] instead.
    app_label = getattr(meta, 'app_label', None)

    if not app_label:
        # This part differs between Django versions.
        # As of Django 1.7, the app framework is used.
        app_label = get_app_label(module)

    model_name = model_name.lower()
    db_table = "{0}{1}_{2}".format(prefix or '', app_label, model_name)
    return truncate_name(db_table, connection.ops.max_name_length())





"""
Optional integration with django-taggit.
"""
from __future__ import absolute_import
from django.contrib.contenttypes.models import ContentType
from django.db import models
from fluent_utils.django_compat import is_installed

__all__ = (
    'TaggableManager',
    'TagsMixin',
)

TaggableManager = None

if is_installed('taggit_autosuggest'):
    from taggit_autosuggest.managers import TaggableManager
elif is_installed('taggit_autocomplete_modified'):
    from taggit_autocomplete_modified.managers import TaggableManagerAutocomplete as TaggableManager
elif is_installed('taggit'):
    from taggit.managers import TaggableManager


# Make sure the 'tags' field is ignored by old versions of South
try:
    from south.modelsinspector import add_ignored_fields
except ImportError:
    pass
else:
    # South should ignore the tags field as it's a RelatedField.
    add_ignored_fields((
        "^taggit\.managers\.TaggableManager",
        "^taggit_autosuggest\.managers\.TaggableManager",
        "^taggit_autocomplete_modified\.managers\.TaggableManagerAutocomplete",
    ))


class TagsMixin(models.Model):
    """
    Mixin for adding tags to a model.
    """
    # Make association with tags optional.
    if TaggableManager is not None:
        tags = TaggableManager(blank=True)
    else:
        tags = None

    class Meta:
        abstract = True


    def similar_objects(self, num=None, **filters):
        """
        Find similar objects using related tags.
        """
        tags = self.tags
        if not tags:
            return []

        content_type = ContentType.objects.get_for_model(self.__class__)
        filters['content_type'] = content_type

        # can't filter, see
        # - https://github.com/alex/django-taggit/issues/32
        # - http://django-taggit.readthedocs.org/en/latest/api.html#TaggableManager.similar_objects
        #
        # Otherwise this would be possible:
        # return tags.similar_objects(**filters)

        lookup_kwargs = tags._lookup_kwargs()
        lookup_keys = sorted(lookup_kwargs)
        qs = tags.through.objects.values(*lookup_kwargs.keys())
        qs = qs.annotate(n=models.Count('pk'))
        qs = qs.exclude(**lookup_kwargs)
        subq = tags.all()
        qs = qs.filter(tag__in=list(subq))
        qs = qs.order_by('-n')

        # from https://github.com/alex/django-taggit/issues/32#issuecomment-1002491
        if filters is not None:
            qs = qs.filter(**filters)

        if num is not None:
            qs = qs[:num]

        # Normal taggit code continues

        # TODO: This all feels like a bit of a hack.
        items = {}
        if len(lookup_keys) == 1:
            # Can we do this without a second query by using a select_related()
            # somehow?
            f = tags.through._meta.get_field_by_name(lookup_keys[0])[0]
            objs = f.rel.to._default_manager.filter(**{
                "%s__in" % f.rel.field_name: [r["content_object"] for r in qs]
            })
            for obj in objs:
                items[(getattr(obj, f.rel.field_name),)] = obj
        else:
            preload = {}
            for result in qs:
                preload.setdefault(result['content_type'], set())
                preload[result["content_type"]].add(result["object_id"])

            for ct, obj_ids in preload.items():
                ct = ContentType.objects.get_for_id(ct)
                for obj in ct.model_class()._default_manager.filter(pk__in=obj_ids):
                    items[(ct.pk, obj.pk)] = obj

        results = []
        for result in qs:
            obj = items[
                tuple(result[k] for k in lookup_keys)
            ]
            obj.similar_tags = result["n"]
            results.append(obj)
        return results



import sys
from django.conf import settings


try:
    from django.apps import apps
except ImportError:
    from django.db.backends.util import truncate_name

    # Django 1.6 or below.
    def is_installed(appname):
        return appname in settings.INSTALLED_APPS

    from django.db.models import get_models

    def get_app_names():
        return settings.INSTALLED_APPS

    def get_app_label(module):
        # Django <= 1.6 does not know apps.
        # Figure out the app_label by looking one level up.
        # For 'django.contrib.sites.models', this would be 'sites'.
        model_module = sys.modules[module]
        return model_module.__name__.split('.')[-2]
else:
    from django.db.backends.utils import truncate_name

    # Django 1.7 provides an official API, and INSTALLED_APPS may contain non-string values too.
    is_installed = apps.is_installed
    get_models = apps.get_models

    def get_app_names():
        return [
            appconfig.name for appconfig in apps.get_app_configs()
        ]

    def get_app_label(module):
        # Django 1.7 knows the "apps", so it can provide the meta information
        app_config = apps.get_containing_app_config(module)
        return app_config.label



from django.db import models


class HideChoicesCharField(models.CharField):
    """
    For Django 1.7, hide the 'choices' for a field.
    """
    def deconstruct(self):
        name, path, args, kwargs = models.CharField.deconstruct(self)

        # Hide the fact this model was used.
        if path == 'fluent_utils.dry.fields.HideChoicesCharField':
            path = 'django.db.models.CharField'
        try:
            del kwargs['choices']
        except KeyError:
            pass

        return name, path, args, kwargs

try:
    from south.modelsinspector import add_introspection_rules
except ImportError:
    pass
else:
    add_introspection_rules([], ["^" + __name__.replace(".", "\.") + "\.HideChoicesCharField"])
